#ifndef WPIPE_FRAMECACHE_HH
#define WPIPE_FRAMECACHE_HH
#include <string>

namespace wpipe {

  /**  Maintain a list of frames and other data sources for use by the 
    *  readdata functions.
    */
  class wframecache {
  public:
    /** GPS time data type
      */
    typedef unsigned long gps_type;

    /**  Read a list of frame types and source directories from a specified 
      *  file. The cache file contains the following columns:
      *  - Observatory ID (\e e.g. "H", "L" or "V")
      *  - Frame Type \e e.g. "R"
      *  - Start gps
      *  - End gps
      *  - File length in seconds
      *  - Frame file directory path
      *
      *  For special data sources (\e e.g. nds) the file length is specified
      *  as zero and the path specifier does not need to be modified.
      *  \brief Data constructor. The 
      *  \param file List file name.
      */
    wframecache(const std::string& file);

    /**  Get a list of frames for the given gps segment and frame type.
      *  \brief Get frame file list.
      *  \param prefix Combined observatory and frame type code. 
      *  \param start Start time of required data.
      *  \param stop  End time of required data.
      *  \param frames String vector to receive frame paths.
      *  \return Number of frame IDs.
      */
    int get_list(const std::string& prefix, gps_type start, 
		 gps_type stop, str_vect& frames) const;
  private:
    class frame_group {
    public:
      frame_group(const std::string& prefix, gps_type start, gps_type stop,
		  gps_type frame_len, const std::string& dir);
      bool valid(const std::string& prefix, gps_type start, 
		 gps_type stop) const;
      std::string frame(gps_type gps) const;
      int get_frames(gps_type start, gps_type stop, str_vect& frames) const;
     private:
      std::string _prefix;
      std::string _directory;
      gps_type _start;
      gps_type _length;
      gps_type _end;
    };
    typedef std::vector<frame_group> cache_list;
    cache_list frameCache;
  };

}  // namespace wpipe
#endif // !defined(WPIPE_FRAMECACHE_HH)
